#! /usr/bin/python

#
# POP3Server.py 0.9
#
# Copyright (C) 2001,2002 Gerson Kurz (not@p-nand-q.com)
# BSD-Licensed. This is free software, enjoy!

"""SHICKS! POP3 Server for Python

This server implements a RFC 1725 compliant POP3 server with a MySQL-based
mailbox storage. The RFC can be found online at

    http://www.ietf.org/rfc/rfc1725.txt

The comments for the commands below are taken from that document. The
database format is discussed in README.HTM. This module is designed to
be reusable in other python projects.
"""

import sys, socket, SocketServer, os, time, ShicksTrace
import socket, md5, rotor, binascii, traceback, thread

CURRENT_VERSION = "0.9"

import ShicksUser # user management
import ShicksMessageList # message database
from ShicksConfig import GetConfigSettings
from POP3StatusMessages import status_msgs

# If this flag is set, deleted messages will be removed from the database, and physically
# deleted from the disk
PHYSICALLY_DELETE_MAILS = int(GetConfigSettings("PhysicallyDeleteMails","0"))
   
class POP3Server(SocketServer.TCPServer):
    allow_reuse_address = 1
    def server_bind(self):
        """Override server_bind to store the server name."""
        SocketServer.TCPServer.server_bind(self)
        host, port = self.socket.getsockname()
        self.server_name = socket.getfqdn(host)
        self.server_port = port

class POP3RequestHandler(SocketServer.ThreadingMixIn, SocketServer.StreamRequestHandler):

    def SendMappedResponse(self,msg,args=None):
        msg = self.status_messages[msg]
        if args is not None:
            msg = msg % args

        print "WRITE:" + msg.strip()
        self.wfile.write(msg)
        self.wfile.flush()

    def SendResponse(self,msg,displaymsg=1):
        if displaymsg:        
            print "WRITE:" + msg.strip()
        else:
            print "(Transmitting mail body)"
        self.wfile.write(msg)
        self.wfile.flush()

    def handle(self):
        # POP3 mailbox state variables        
        self.user = None
        self.transaction_state = 0
        self.messagelist = None
        self.terminated = 0
        self.status_messages = status_msgs

        # supported commands: all required and optional commands as mentioned in said RFC
        cmdmap = { 'QUIT' : self.QuitConnection,
                   'USER' : self.SetUserName,
                   'PASS' : self.SetUserPassword,
                   'LIST' : self.ListMessages,
                   'UIDL' : self.GetUniqueIDList,
                   'DELE' : self.OnDeleteMessage,
                   'STAT' : self.GetStatistics,
                   'RSET' : self.ResetDeletedMessageIDs,
                   'RETR' : self.OnRetrieveMessage,
                   'APOP' : self.Md5SecureLogon,
                   'TOP'  : self.GetMessageTopLines }

        print "BEGIN of POP3 conversation with " + str(self.client_address)
        self.secret = "<%d.%d.%s@%s.%s>" % (os.getpid(),thread.get_ident(),str(time.clock()),socket.gethostname(),str(id(self)))
        self.SendMappedResponse('OK_GREETING', (CURRENT_VERSION, self.secret))

        while not self.terminated:
            self.inputline = self.rfile.readline().strip()

            if self.inputline[:5].upper() in ('PASS:','PASS '):
                print "READ: PASS: *******"
            else:
                print "READ:" + self.inputline.strip()
            request = self.inputline.split()
            if len(request):
                request[0] = request[0].upper()
                if cmdmap.has_key(request[0]):
                    cmdmap[request[0]](request)
                else:
                    self.SendMappedResponse('ERR_CMD_UNKNOWN')
            else:
                break
        print "END of POP3 conversation with " + str(self.client_address)

    def SetUserName(self,request):
        """USER name

         Arguments:
             a string identifying a mailbox (required), which is of
             significance ONLY to the server

         Restrictions:
             may only be given in the AUTHORIZATION state after the POP3
             greeting or after an unsuccessful USER or PASS command

         Possible Responses:
             +OK name is a valid mailbox
             -ERR never heard of mailbox name

         Examples:
             C: USER mrose
             S: +OK mrose is a real hoopy frood
                ...
             C: USER frated
             S: -ERR sorry, no mailbox for frated here"""

        if self.transaction_state:
            return self.SendMappedResponse('ERR_INVALID_STATE')

        self.user = ShicksUser.GetUserByName(request[1])
        if self.user:
            self.SendMappedResponse('OK_USER',request[1])
        else:
            self.SendMappedResponse('ERR_NOT_USER')

    def SetUserPassword(self,request):
        """PASS string

         Arguments:
             a server/mailbox-specific password (required)

         Restrictions:
             may only be given in the AUTHORIZATION state after a
             successful USER command

         Discussion:
             Since the PASS command has exactly one argument, a POP3
             server may treat spaces in the argument as part of the
             password, instead of as argument separators.

         Possible Responses:
             +OK maildrop locked and ready
             -ERR invalid password
             -ERR unable to lock maildrop

         Examples:
             C: USER mrose
             S: +OK mrose is a real hoopy frood
             C: PASS secret
             S: +OK mrose's maildrop has 2 messages (320 octets)
               ...
             C: USER mrose
             S: +OK mrose is a real hoopy frood
             C: PASS secret
             S: -ERR maildrop already locked"""

        if self.transaction_state:
            return self.SendMappedResponse('ERR_INVALID_STATE')

        if self.user is None:
            return self.SendMappedResponse('LOGIN_REQUIRED')
            
        pwd = ""
        if len(request) > 1:
            pwd = self.inputline[5:]

        if pwd == self.user.getPassword():
            self.GetMessagesForUserID()
            print "Have %d message(s) in mailbox..." % len(self.messagelist)
            self.SendMappedResponse('OK_LOGIN')
            self.transaction_state = 1
        else:
            print "ERROR, invalid password!"
            self.SendMappedResponse('ERR_LOGIN_MISMATCH')

    def ListMessages(self,request):
        """LIST [msg]

         Arguments:
             a message-number (optional), which, if present, may NOT
             refer to a message marked as deleted

         Restrictions:
             may only be given in the TRANSACTION state

         Discussion:
             If an argument was given and the POP3 server issues a
             positive response with a line containing information for
             that message.  This line is called a "scan listing" for
             that message.

             If no argument was given and the POP3 server issues a
             positive response, then the response given is multi-line.

             After the initial +OK, for each message in the maildrop,
             the POP3 server responds with a line containing information
             for that message.  This line is also called a "scan
             listing" for that message.

             In order to simplify parsing, all POP3 servers are required
             to use a certain format for scan listings.  A scan listing
             consists of the message-number of the message, followed by
             a single space and the exact size of the message in octets.
             This memo makes no requirement on what follows the message
             size in the scan listing.  Minimal implementations should
             just end that line of the response with a CRLF pair.  

             Note that messages marked as deleted are not listed.

         Possible Responses:
             +OK scan listing follows
             -ERR no such message

         Examples:
             C: LIST
             S: +OK 2 messages (320 octets)
             S: 1 120
             S: 2 200
             S: .
               ...
             C: LIST 2
             S: +OK 2 200
               ...
             C: LIST 3
             S: -ERR no such message, only 2 messages in maildrop"""
        
        if not self.transaction_state:
            return self.SendMappedResponse('ERR_INVALID_STATE')

        # handle optional message-id parameter
        messageid = None
        if len(request) > 1:
            try:
                messageid = int(request[1])
            except:
                print "Warning, LIST without proper request ID: " + str(request)
                messageid = -1

        if not messageid:
            usedbytes = 0
            response = ""
            index = 0
            for message in self.messagelist:
                index += 1
                
                # skip deleted messages
                if message.getDeleted():
                    continue

                response += "%d %d\r\n" % (index,message.getMessageSize())
                usedbytes += message.getMessageSize()
            response = "+OK %d %d\r\n" % (len(self.messagelist),usedbytes)+response+".\r\n"
        else:
            response = None
            index = 0
            for message in self.messagelist:
                index += 1
                if messageid == index:
                    response = "+OK %d %d\r\n" % (index,message.getMessageSize())
                    break
                
            if not response:
                return self.SendMappedResponse('ERR_MSG_UNKNOWN')
            
        self.SendResponse(response)

    def GetUniqueIDList(self,request):
        """UIDL [msg]

      Arguments:
          a message-number (optionally)  If a message-number is given,
          it may NOT refer to a message marked as deleted.

      Restrictions:
          may only be given in the TRANSACTION state.

      Discussion:
          If an argument was given and the POP3 server issues a positive
          response with a line containing information for that message.
          This line is called a "unique-id listing" for that message.

          If no argument was given and the POP3 server issues a positive
          response, then the response given is multi-line.  After the
          initial +OK, for each message in the maildrop, the POP3 server
          responds with a line containing information for that message.
          This line is called a "unique-id listing" for that message.

          In order to simplify parsing, all POP3 servers are required to
          use a certain format for unique-id listings.  A unique-id
          listing consists of the message-number of the message,
          followed by a single space and the unique-id of the message.

          No information follows the unique-id in the unique-id listing.

          The unique-id of a message is an arbitrary server-determined
          string, consisting of characters in the range 0x21 to 0x7E,
          which uniquely identifies a message within a maildrop and
          which persists across sessions. The server should never reuse
          an unique-id in a given maildrop, for as long as the entity
          using the unique-id exists.

          Note that messages marked as deleted are not listed.

      Possible Responses:
          +OK unique-id listing follows
          -ERR no such message

      Examples:
          C: UIDL
          S: +OK
          S: 1 whqtswO00WBw418f9t5JxYwZ
          S: 2 QhdPYR:00WBw1Ph7x7
          S: .
             ...
          C: UIDL 2
          S: +OK 2 QhdPYR:00WBw1Ph7x7
             ...
          C: UIDL 3
          S: -ERR no such message, only 2 messages in maildrop"""

        if not self.transaction_state:
            return self.SendMappedResponse('ERR_INVALID_STATE')

        # handle optional message-id parameter
        messageid = None
        if len(request) > 1:
            try:
                messageid = int(request[1])
            except:
                print "Warning, UIDL without proper request ID: " + str(request)
                messageid = -1

        # create response
        if messageid is None:
            response = "+OK\r\n"
            index = 0
            for message in self.messagelist:
                index += 1

                # skip deleted messages
                if message.getDeleted():
                    continue

                response += "%d shicks%d.%s.%d\r\n" % (index,os.getpid(),str(time.clock()),message.getId())
            response += ".\r\n"
        else:
            response = None
            index = 0
            for message in self.messagelist:
                index += 1
                if messageid == index:
                    response = "+OK %d shicks%d.%s.%d\r\n" % (index,os.getpid(),str(time.clock()),messag.getId())
                    break
                
            if not response:
                return self.SendMappedResponse('ERR_MSG_UNKNOWN')
            
        self.SendResponse(response)

    def OnRetrieveMessage(self,request):
        """RETR msg

         Arguments:
             a message-number (required) which may not refer to a
             message marked as deleted

         Restrictions:
             may only be given in the TRANSACTION state

         Discussion:
             If the POP3 server issues a positive response, then the
             response given is multi-line.  After the initial +OK, the
             POP3 server sends the message corresponding to the given
             message-number, being careful to byte-stuff the termination
             character (as with all multi-line responses).

         Possible Responses:
             +OK message follows
             -ERR no such message

         Examples:
             C: RETR 1
             S: +OK 120 octets
             S: <the POP3 server sends the entire message here>
             S: ."""
        
        if not self.transaction_state:
            return self.SendMappedResponse('ERR_INVALID_STATE')

        try:
            messageid = int(request[1])
        except:
            print "Warning, RETR without proper request ID:", request
            messageid = -1
        
        body = self.GetEnvelopedMessageBody(messageid)
        if body is None:
            self.SendMappedResponse('ERR_MSG_UNKNOWN')
        else:
            self.SendResponse("+OK message follows\r\n"+body+"\r\n.\r\n",0)

    def OnDeleteMessage(self,request):
        """DELE msg

         Arguments:
             a message-number (required) which may not refer to a
             message marked as deleted

         Restrictions:
             may only be given in the TRANSACTION state

         Discussion:
             The POP3 server marks the message as deleted.  Any future
             reference to the message-number associated with the message
             in a POP3 command generates an error. The POP3 server does
             not actually delete the message until the POP3 session enters
             the UPDATE state.

         Possible Responses:
             +OK message deleted
             -ERR no such message

         Examples:
             C: DELE 1
             S: +OK message 1 deleted
                ...
             C: DELE 2
             S: -ERR message 2 already deleted"""
        
        if not self.transaction_state:
            return self.SendMappedResponse('ERR_INVALID_STATE')

        try:
            messageid = int(request[1])
        except:
            return self.SendMappedResponse('ERR_MSG_UNKNOWN')

        if messageid >= 1 and messageid <= len(self.messagelist):
            messageid -= 1
            # mark message as deleted
            self.messagelist[messageid].setDeleted(1)
            return self.SendMappedResponse('OK_DELETE')

        self.SendMappedResponse('ERR_MSG_UNKNOWN')

    def GetStatistics(self,request):
        """STAT

         Arguments: none

         Restrictions:
             may only be given in the TRANSACTION state

         Discussion:
             The POP3 server issues a positive response with a line
             containing information for the maildrop.  This line is
             called a "drop listing" for that maildrop.

             In order to simplify parsing, all POP3 servers required to
             use a certain format for drop listings.  The positive
             response consists of "+OK" followed by a single space, the
             number of messages in the maildrop, a single space, and the
             size of the maildrop in octets.  This memo makes no
             requirement on what follows the maildrop size.  Minimal
             implementations should just end that line of the response
             with a CRLF pair.  More advanced implementations may
             include other information.

             Note that messages marked as deleted are not counted in
             either total.

         Possible Responses:
             +OK nn mm

         Examples:
             C: STAT
             S: +OK 2 320"""
        
        if not self.transaction_state:
            return self.SendMappedResponse('ERR_INVALID_STATE')

        usedbytes = 0
        response = ""
        for message in self.messagelist:

            # skip deleted messages
            if message.getDeleted():
                continue

            usedbytes += message.getMessageSize()
        self.SendMappedResponse('OK_SINGLE_LIST',(len(self.messagelist),usedbytes))
        
    def ResetDeletedMessageIDs(self,request):
        """RSET

         Arguments: none

         Restrictions:
             may only be given in the TRANSACTION state

         Discussion:
             If any messages have been marked as deleted by the POP3
             server, they are unmarked.  The POP3 server then replies
             with a positive response.

         Possible Responses:
             +OK

         Examples:
             C: RSET
             S: +OK maildrop has 2 messages (320 octets)"""

        if not self.transaction_state:
            return self.SendMapppedResponse('ERR_INVALID_STATE')
    
        for i in xrange(len(self.messagelist)):
            # mark all messages as not-deleted 
            self.messagelist[i].setDeleted(0)

        self.SendMappedResponse('OK_RSET')       

    def GetMessageTopLines(self,request):
        """TOP msg n

         Arguments:
             a message-number (required) which may NOT refer to to a
             message marked as deleted, and a non-negative number
             (required)

         Restrictions:
             may only be given in the TRANSACTION state

         Discussion:
             If the POP3 server issues a positive response, then the
             response given is multi-line.  After the initial +OK, the
             POP3 server sends the headers of the message, the blank
             line separating the headers from the body, and then the
             number of lines indicated message's body, being careful to
             byte-stuff the termination character (as with all multi-
             line responses).

             Note that if the number of lines requested by the POP3
             client is greater than than the number of lines in the
             body, then the POP3 server sends the entire message.

         Possible Responses:
             +OK top of message follows
             -ERR no such message

         Examples:
             C: TOP 1 10
             S: +OK
             S: <the POP3 server sends the headers of the
                message, a blank line, and the first 10 lines
                of the body of the message>
             S: .
                ...
             C: TOP 100 3
             S: -ERR no such message"""

        if not self.transaction_state:
            return self.SendMappedResponse('ERR_INVALID_STATE')

        if len(request) != 3:
            return self.SendMappedResponse('ERR_INVALID_SYNTAX')

        try:
            messageid = int(request[1])
        except:
            print "Warning, TOP without proper request ID:", request
            messageid = -1

        try:
            nroflines = int(request[2])
        except:
            print "Warning, TOP without proper request ID:", request
            nroflines = 0
        
        body = self.GetEnvelopedMessageBody(messageid)
        print "read body", len(body)
        if body is not None:
            lines = body.split("\r\n")

            found, isHeader, line = -1, 1, 0
            while line < len(lines):
                if isHeader:
                    # skip header lines. Header lines have a :
                    if lines[line].find(':') == 0:
                        isHeader = 0
                else:
                    found += 1
                    if found == nroflines:
                        lines = lines[:line]
                        body = "\r\n".join(lines)
                        break
                line += 1
            self.SendResponse("+OK\r\n"+body+"\r\n.\r\n")
        else:                    
            self.SendMappedResponse('ERR_MSG_UNKNOWN')
             
    def QuitConnection(self,request):
        """QUIT

         Arguments: none

         Restrictions: none

         Discussion:
             The POP3 server removes all messages marked as deleted from
             the maildrop.  It then releases any exclusive-access lock
             on the maildrop and replies as to the status of these
             operations.  The TCP connection is then closed.

         Possible Responses:
             +OK

         Examples:
             C: QUIT
             S: +OK dewey POP3 server signing off (maildrop empty)
                ...
             C: QUIT
             S: +OK dewey POP3 server signing off (2 messages left)
                ..."""

        self.terminated = 1
        self.SendMappedResponse('OK_QUIT')

        ids = []
        for message in self.messagelist:
            if message.getDeleted():
                ids.append(message.getId())

        if len(ids):
            ShicksMessageList.DeleteMessageBodies(ids,PHYSICALLY_DELETE_MAILS)

    def Md5SecureLogon(self,request):
        """APOP name digest

         Arguments:
             a string identifying a mailbox and a MD5 digest string
             (both required)

         Restrictions:
             may only be given in the AUTHORIZATION state after the POP3
             greeting

         Discussion:
             Normally, each POP3 session starts with a USER/PASS
             exchange.  This results in a server/user-id specific
             password being sent in the clear on the network.  For
             intermittent use of POP3, this may not introduce a sizable
             risk.  However, many POP3 client implementations connect to
             the POP3 server on a regular basis -- to check for new
             mail.  Further the interval of session initiation may be on
             the order of five minutes.  Hence, the risk of password
             capture is greatly enhanced.

             An alternate method of authentication is required which
             provides for both origin authentication and replay
             protection, but which does not involve sending a password
             in the clear over the network.  The APOP command provides
             this functionality.

             A POP3 server which implements the APOP command will
             include a timestamp in its banner greeting.  The syntax of
             the timestamp corresponds to the `msg-id' in [RFC822], and
             MUST be different each time the POP3 server issues a banner
             greeting.  For example, on a UNIX implementation in which a
             separate UNIX process is used for each instance of a POP3
             server, the syntax of the timestamp might be:

                <process-ID.clock@hostname>

             where `process-ID' is the decimal value of the process's
             PID, clock is the decimal value of the system clock, and
             hostname is the fully-qualified domain-name corresponding
             to the host where the POP3 server is running.

             The POP3 client makes note of this timestamp, and then
             issues the APOP command.  The `name' parameter has
             identical semantics to the `name' parameter of the USER
             command. The `digest' parameter is calculated by applying
             the MD5 algorithm [RFC1321] to a string consisting of the
             timestamp (including angle-brackets) followed by a shared
             secret.  This shared secret is a string known only to the
             POP3 client and server.  Great care should be taken to
             prevent unauthorized disclosure of the secret, as knowledge
             of the secret will allow any entity to successfully
             masquerade as the named user.  The `digest' parameter
             itself is a 16-octet value which is sent in hexadecimal
             format, using lower-case ASCII characters.

             When the POP3 server receives the APOP command, it verifies
             the digest provided.  If the digest is correct, the POP3
             server issues a positive response, and the POP3 session
             enters the TRANSACTION state.  Otherwise, a negative
             response is issued and the POP3 session remains in the
             AUTHORIZATION state.

             Note that as the length of the shared secret increases, so
             does the difficulty of deriving it.  As such, shared
             secrets should be long strings (considerably longer than
             the 8-character example shown below).

         Possible Responses:
             +OK maildrop locked and ready
             -ERR permission denied

         Examples:
             S: +OK POP3 server ready <1896.697170952@dbc.mtview.ca.us>
             C: APOP mrose c4c9334bac560ecc979e58001b3e22fb
             S: +OK maildrop has 1 message (369 octets)

             In this example, the shared  secret  is  the  string  `tan-
             staaf'.  Hence, the MD5 algorithm is applied to the string

                <1896.697170952@dbc.mtview.ca.us>tanstaaf

             which produces a digest value of

                c4c9334bac560ecc979e58001b3e22fb"""
        
        if self.transaction_state:
            return self.SendMappedResponse('ERR_INVALID_STATE')

        if len(request) != 3:
            return self.SendMappedResponse('ERR_INVALID_SYNTAX')

        self.user = ShicksUser.GetUserByName(request[1])
        if not self.user:
            return self.SendMappedResponse('ERR_NOT_USER')

        expected = md5.new(self.secret + self.user.getPassword()).hexdigest()
        if expected != request[2]:
            print "ERROR, invalid password: expected",expected,"but got",request[2]
            return self.SendMappedResponse('ERR_LOGIN_MISMATCH')
        
        self.GetMessagesForUserID()
        print "Have %d message(s) in mailbox..." % len(self.messagelist)
        self.SendMappedResponse('OK_APOP')
        self.transaction_state = 1
       
    def GetMessagesForUserID(self):
        self.messagelist = ShicksMessageList.GetUnreadMessagesForUserID(self.user.getId())

    def GetEnvelopedMessageBody(self,messageid):
        if messageid >= 1 and messageid <= len(self.messagelist):
            message = self.messagelist[messageid-1]
        else:
            print "ERROR, invalid message id", messageid, "only have", len(self.messagelist), "messages"
            return None
        
        body = message.getFilename()
        if body:
            result = None
            try:
                file = open(body,"rb")
                result = file.read()
                file.close()
            except:
                traceback.print_exc()
            return result

def StartPOP3Server():
    server = POP3Server(('', 110), POP3RequestHandler)
    print "POP3 server starting - Version %s on %s" % (CURRENT_VERSION,socket.gethostname())
    server.serve_forever()
    print "POP3 server returned"

if __name__ == '__main__':
    TRACE_DIRECTORY = GetConfigSettings("Logfiles",(os.name == "nt") and "c:\\shicks\\log" or "/usr/share/shicks/log")
    try:
        os.makedirs(TRACE_DIRECTORY )
    except:
        pass
    ShicksTrace.TraceFileName = os.path.join(TRACE_DIRECTORY,"POP3Server.log")
    ShicksTrace.Enable()    
    StartPOP3Server()
    
